import os
from pkg_resources import resource_filename
from mot.kernel_data import KernelAllocatedArray
from .base import AbstractParallelOptimizer

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LevenbergMarquardt(AbstractParallelOptimizer):

    default_patience = 250

    def __init__(self, patience=None, step_bound=None, scale_diag=None, optimizer_settings=None, **kwargs):
        """Use the Levenberg-Marquardt method to calculate the optimimum.

        Args:
            patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}

        keyword_values = {}
        keyword_values['step_bound'] = step_bound
        keyword_values['scale_diag'] = scale_diag

        option_defaults = {'step_bound': 100.0, 'scale_diag': 1}

        def get_value(option_name):
            value = keyword_values.get(option_name)
            if value is None:
                value = optimizer_settings.get(option_name)
            if value is None:
                value = option_defaults[option_name]
            return value

        for option in option_defaults:
            optimizer_settings.update({option: get_value(option)})

        super(LevenbergMarquardt, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def minimize(self, model, starting_positions):
        if model.get_nmr_observations() < model.get_nmr_parameters():
            raise ValueError('The number of instances per problem must be greater than the number of parameters')
        return super(LevenbergMarquardt, self).minimize(model, starting_positions)

    def _get_optimizer_kernel_data(self, model):
        return {'_fjac_all': KernelAllocatedArray((model.get_nmr_problems(),
                                                   model.get_nmr_parameters(),
                                                   model.get_nmr_observations()), ctype='mot_float_type',
                                                  is_writable=True, is_readable=True)}

    def _get_optimizer_call_args(self):
        return super(LevenbergMarquardt, self)._get_optimizer_call_args() + ['data->_fjac_all']

    def _get_evaluate_function(self, model):
        """Get the CL code for the evaluation function. This is called from _get_optimizer_cl_code.

        Implementing optimizers can change this if desired.

        Returns:
            str: the evaluation function.
        """
        objective_func = model.get_objective_per_observation_function()
        param_modifier = model.get_pre_eval_parameter_modifier()

        kernel_source = ''
        kernel_source += objective_func.get_cl_code()
        kernel_source += param_modifier.get_cl_code()
        kernel_source += '''
            void evaluate(mot_float_type* x, void* data_void, mot_float_type* result){
                mot_data_struct* data = (mot_data_struct*)data_void;
                
                mot_float_type x_model[''' + str(model.get_nmr_parameters()) + '''];
                for(uint i = 0; i < ''' + str(model.get_nmr_parameters()) + '''; i++){
                    x_model[i] = x[i];
                }
                ''' + param_modifier.get_cl_function_name() + '''(data, x_model);
                
                for(uint i = 0; i < ''' + str(model.get_nmr_observations()) + '''; i++){
                    // the model expects the L1 norm, while the LM method takes the L2 norm. Taking square root here.
                    result[i] = sqrt(fabs(''' + objective_func.get_cl_function_name() + '''(data, x_model, i)));
                }
            }
        '''
        return kernel_source

    def _get_optimization_function(self, model):
        params = {'NMR_PARAMS': model.get_nmr_parameters(),
                  'PATIENCE': self.patience,
                  'NMR_OBSERVATIONS': model.get_nmr_observations(),
                  'USER_TOL_MULT': 30}

        optimizer_settings = self._optimizer_settings or {}
        option_defaults = {'step_bound': 100.0, 'scale_diag': 1, 'usertol_mult': 30}
        option_converters = {'scale_diag': lambda val: int(bool(val))}

        for option, default in option_defaults.items():
            v = optimizer_settings.get(option, default)
            if option in option_converters:
                v = option_converters[option](v)
            params.update({option.upper(): v})

        body = open(os.path.abspath(resource_filename('mot', 'data/opencl/lmmin.cl')), 'r').read()
        if params:
            body = body % params
        return body

    def _get_optimizer_call_name(self):
        return 'lmmin'



import logging
import numpy as np

from mot.cl_routines.mapping.objective_function_calculator import ObjectiveFunctionCalculator
from mot.cl_routines.mapping.run_procedure import RunProcedure
from mot.kernel_data import KernelArray, KernelAllocatedArray
from ...utils import get_float_type_def, NameFunctionTuple
from ...cl_routines.base import CLRoutine
from ...__version__ import __version__

__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


return_code_labels = {
    0: ['default', 'no return code specified'],
    1: ['found zero', 'sum of squares/evaluation below underflow limit'],
    2: ['converged', 'the relative error in the sum of squares/evaluation is at most tol'],
    3: ['converged', 'the relative error of the parameter vector is at most tol'],
    4: ['converged', 'both errors are at most tol'],
    5: ['trapped', 'by degeneracy; increasing epsilon might help'],
    6: ['exhausted', 'number of function calls exceeding preset patience'],
    7: ['failed', 'ftol<tol: cannot reduce sum of squares any further'],
    8: ['failed', 'xtol<tol: cannot improve approximate solution any further'],
    9: ['failed', 'gtol<tol: cannot improve approximate solution any further'],
    10: ['NaN', 'Function value is not-a-number or infinite'],
    11: ['exhausted', 'temperature decreased to 0.0']
}


class AbstractOptimizer(CLRoutine):

    def __init__(self, patience=1, optimizer_settings=None, **kwargs):
        """Create a new optimizer that will minimize the given model using the given environments.

        If the environment is None, a suitable default environment is created.

        Args:
            patience (int): The patience is used in the calculation of how many iterations to iterate the optimizer.
                The exact usage of this value of this parameter may change per optimizer.
            optimizer_options (dict): extra options one can set for the optimization routine. These are routine
                dependent.
        """
        super(AbstractOptimizer, self).__init__(**kwargs)

        self._optimizer_settings = optimizer_settings or {}
        if not isinstance(self._optimizer_settings, dict):
            raise ValueError('The given optimizer settings is not a dictionary.')

        self.patience = patience or 1

        if 'patience' in self._optimizer_settings:
            self.patience = self._optimizer_settings['patience']
        self._optimizer_settings['patience'] = self.patience

    @property
    def optimizer_settings(self):
        """Get the optimization flags and settings set to the optimizer.

        This ordinarily should contain all the extra flags and options set to the optimization routine. Even those
        set additionally in the constructor.

        Returns:
            dict: the optimization options (settings)
        """
        return self._optimizer_settings

    def minimize(self, model, starting_positions):
        """Minimize the given model using the given environments.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): The model to minimize
            starting_positions (ndarray): The starting starting_positions for every problem in the model, it should be
                an matrix of shape (d, p) with d problems and p parameter.

        Returns:
            OptimizationResults: the container with the optimization results
        """
        raise NotImplementedError()


class OptimizationResults(object):

    def get_optimization_result(self):
        """Get the optimization result, that is, the matrix with the optimized parameter values.

        Returns:
            ndarray: the optimized parameter maps, an (d, p) array with for d problems a value for every p parameters
        """
        raise NotImplementedError()

    def get_return_codes(self):
        """Get the return codes, that is, the matrix with for every problem the return code of the optimizer

        Returns:
            ndarray: the return codes, an (d,) vector with for d problems a return code
        """
        raise NotImplementedError()

    def get_objective_values(self):
        """Get the objective values for each of the problem instances.

        Returns:
            ndarray: (d,) matrix with for every problem d, the objective value
        """
        raise NotImplementedError()


class SimpleOptimizationResult(OptimizationResults):

    def __init__(self, model, optimization_results, return_codes):
        """Simple optimization results container which computes some values only when requested.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): the model we used to get these results
            optimization_results (ndarray): a (d, p) matrix with for every d problems the estimated value for
                every parameter p
            return_codes (ndarray): the return codes as a (d,) vector for every d problems
        """
        self._model = model
        self._optimization_results = optimization_results
        self._return_codes = return_codes
        self._objective_values = None

    def get_optimization_result(self):
        return self._optimization_results

    def get_return_codes(self):
        return self._return_codes

    def get_objective_values(self):
        if self._objective_values is None:
            self._objective_values = np.nan_to_num(ObjectiveFunctionCalculator().calculate(
                self._model, self._optimization_results))
        return self._objective_values


class AbstractParallelOptimizer(AbstractOptimizer):

    def __init__(self, **kwargs):
        """
        Args:
            optimizer_settings (dict): extra options one can set for the optimization routine. These are routine
                dependent.
        """
        super(AbstractParallelOptimizer, self).__init__(**kwargs)
        self._logger = logging.getLogger(__name__)

    def minimize(self, model, starting_positions):
        if starting_positions.shape[0] != model.get_nmr_problems():
            raise ValueError('The number of problems in the model does not match the number of starting points given.')

        if starting_positions.shape[1] != model.get_nmr_parameters():
            raise ValueError('The number of parameters in the model does not match the number of '
                             'starting points given.')

        self._logger.info('Entered optimization routine.')
        self._logger.info('Using MOT version {}'.format(__version__))
        self._logger.info('We will use a {} precision float type for the calculations.'.format(
            'double' if self._cl_runtime_info.double_precision else 'single'))
        for env in self._cl_runtime_info.get_cl_environments():
            self._logger.info('Using device \'{}\'.'.format(str(env)))
        self._logger.info('Using compile flags: {}'.format(self._cl_runtime_info.get_compile_flags()))
        self._logger.info('We will use the optimizer {} '
                          'with optimizer settings {}'.format(self.__class__.__name__,
                                                              self._optimizer_settings))

        self._logger.info('Starting optimization preliminaries')

        if len(starting_positions.shape) < 2:
            starting_positions = starting_positions[..., None]

        all_kernel_data = dict(model.get_kernel_data())
        all_kernel_data.update({
            '_parameters': KernelArray(starting_positions, ctype='mot_float_type', is_writable=True, is_readable=True),
            '_return_codes': KernelAllocatedArray((model.get_nmr_problems(),), ctype='char',
                                                  is_readable=False, is_writable=True)
        })
        all_kernel_data.update(self._get_optimizer_kernel_data(model))

        self._logger.info('Finished optimization preliminaries')
        self._logger.info('Starting optimization')

        runner = RunProcedure(self._cl_runtime_info)
        runner.run_procedure(self._get_optimizer_function(model),
                             all_kernel_data, model.get_nmr_problems(),
                             use_local_reduction=False)
        parameters = model.finalize_optimized_parameters(all_kernel_data['_parameters'].get_data())

        self._logger.info('Finished optimization')
        return SimpleOptimizationResult(model, parameters, all_kernel_data['_return_codes'].get_data())

    def _get_optimizer_kernel_data(self, model):
        """Get the kernel data specific to the optimization routines.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): the model we are optimizing

        Returns:
            dict: kernel input data elements
        """
        return {}

    def _get_optimizer_function(self, model):
        """Get the optimization kernel function.

        By default this returns a full kernel source using information from _get_optimizer_cl_code()
        and _get_optimizer_call_name().

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): the model we are optimizing

        Returns:
            NameFunctionTuple: the optimization function to apply
        """
        nmr_params = model.get_nmr_parameters()

        kernel_source = ''
        kernel_source += get_float_type_def(self._cl_runtime_info.double_precision)

        kernel_source += self._get_optimizer_cl_code(model)
        kernel_source += '''
            void compute(mot_data_struct* data){
                mot_float_type x[''' + str(nmr_params) + '''];
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    x[i] = data->_parameters[i];
                }
                
                *data->_return_codes = (char) ''' + self._get_optimizer_call_name() + '(' + \
                         ', '.join(self._get_optimizer_call_args()) + ''');
                
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    data->_parameters[i] = x[i];
                }   
            }
        '''
        return NameFunctionTuple('compute', kernel_source)

    def _get_optimizer_call_args(self):
        """Get the optimizer calling arguments.

        This is useful if a subclass extended the buffers with an additional buffer.

        Returns:
            list of str: the list of optimizer call arguments
        """
        return ['x', '(void*)data']

    def _get_optimizer_cl_code(self, model):
        """Get the optimization CL code that is called during optimization for each problem.

        This is normally called by the default implementation of _get_ll_calculating_kernel().

        By default this creates a CL function named 'evaluation' that can be called by the optimization routine.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): the model we are optimizing

        Returns:
            str: The kernel source for the optimization routine.
        """
        kernel_source = self._get_evaluate_function(model)
        kernel_source += self._get_optimization_function(model)
        return kernel_source

    def _get_evaluate_function(self, model):
        """Get the CL code for the evaluation function. This is called from _get_optimizer_cl_code.

        Implementing optimizers can change this if desired.

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): the model we are optimizing

        Returns:
            str: the evaluation function.
        """
        nmr_params = model.get_nmr_parameters()
        nmr_observations = model.get_nmr_observations()

        objective_function = model.get_objective_per_observation_function()
        param_modifier = model.get_pre_eval_parameter_modifier()

        kernel_source = ''
        kernel_source += objective_function.get_cl_code()
        kernel_source += param_modifier.get_cl_code()
        kernel_source += '''
            double evaluate(mot_float_type* x, void* data_void){

                mot_data_struct* data = (mot_data_struct*)data_void;

                mot_float_type x_model[''' + str(nmr_params) + '''];
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    x_model[i] = x[i];
                }

                ''' + param_modifier.get_cl_function_name() + '''(data, x_model);

                double sum = 0;
                for(uint i = 0; i < ''' + str(nmr_observations) + '''; i++){
                    sum += ''' + objective_function.get_cl_function_name() + '''(data, x_model, i);
                }
                return sum;
            }
        '''
        return kernel_source

    def _get_optimization_function(self, model):
        """Return the optimization function as a CL string for the implementing optimizer.

        This is a convenience function to avoid boilerplate in implementing _get_optimizer_cl_code().

        Args:
            model (mot.model_interfaces.OptimizeModelInterface): the model we are optimizing

        Returns:
            str: The optimization routine function
        """
        raise NotImplementedError()

    def _get_optimizer_call_name(self):
        """Get the call name of the optimization routine.

        This name is the name of the function called by the kernel to optimize a single problem.

        Returns:
            str: The function name of the optimization function.
        """
        raise NotImplementedError()



import numpy as np
from mot.cl_routines.base import CLRoutine
from mot.cl_routines.mapping.run_procedure import RunProcedure
from mot.library_functions import Rand123
from mot.utils import NameFunctionTuple, is_scalar
from mot.kernel_data import KernelArray, KernelAllocatedArray

__author__ = 'Robbert Harms'
__date__ = "2014-10-29"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Random123Generator(CLRoutine):

    def __init__(self, **kwargs):
        """Create the random123 basis for generating multiple lists of random numbers.

        *From the Random123 documentation:*

        Unlike conventional RNGs, counter-based RNGs are stateless functions (or function classes i.e. functors)
        whose arguments are a counter, and a key and returns a result of the same type as the counter.

        .. code-block:: c

            result = CBRNGname(counter, key)

        The returned result is a deterministic function of the key and counter, i.e. a unique (counter, key)
        tuple will always produce the same result. The result is highly sensitive to small changes in the inputs,
        so that the sequence of values produced by simply incrementing the counter (or key) is effectively
        indistinguishable from a sequence of samples of a uniformly distributed random variable.

        All the Random123 generators are counter-based RNGs that use integer multiplication, xor and permutation
        of W-bit words to scramble its N-word input key.

        In this implementation we generate a counter and key automatically from a single seed.

        Args:
            seed (float): the seed, if not given a random seed is used.
        """
        super(Random123Generator, self).__init__(**kwargs)

    def rand(self, nmr_distributions, nmr_samples, min_val=0, max_val=1, ctype='float'):
        """Draw random samples from the Uniform distribution.

        Args:
            nmr_distributions (int): the number of unique continuous_distributions to create
            nmr_samples (int): The number of samples to draw
            minimum (double): The minimum value of the random numbers
            maximum (double): The minimum value of the random numbers
            ctype (str): the C type of the output samples

        Returns:
            ndarray: A two dimensional numpy array as (nmr_distributions, nmr_samples).
        """
        if is_scalar(min_val):
            min_val = np.ones((nmr_distributions, 1)) * min_val
        if is_scalar(max_val):
            max_val = np.ones((nmr_distributions, 1)) * max_val

        kernel_data = {'min_val': KernelArray(min_val),
                       'max_val': KernelArray(max_val)}

        return self._generate_samples(nmr_distributions, nmr_samples, ctype, kernel_data,
                                      self._get_uniform_kernel(nmr_samples, ctype))

    def randn(self, nmr_distributions, nmr_samples, mean=0, std=1, ctype='float'):
        """Draw random samples from the Gaussian distribution.

        Args:
            nmr_distributions (int): the number of unique continuous_distributions to create
            nmr_samples (int): The number of samples to draw
            mean (float or ndarray): The mean of the distribution
            std (float or ndarray): The standard deviation or the distribution
            ctype (str): the C type of the output samples

        Returns:
            ndarray: A two dimensional numpy array as (nmr_distributions, nmr_samples).
        """
        if is_scalar(mean):
            mean = np.ones((nmr_distributions, 1)) * mean
        if is_scalar(std):
            std = np.ones((nmr_distributions, 1)) * std

        kernel_data = {'mean': KernelArray(mean),
                       'std': KernelArray(std)}

        return self._generate_samples(nmr_distributions, nmr_samples, ctype, kernel_data,
                                      self._get_gaussian_kernel(nmr_samples, ctype))

    def _generate_samples(self, nmr_distributions, nmr_samples, ctype, kernel_data, cl_function):
        rng_state = np.random.uniform(low=np.iinfo(np.uint32).min, high=np.iinfo(np.uint32).max + 1,
                                      size=(nmr_distributions, 6)).astype(np.uint32)

        kernel_data.update({'samples': KernelAllocatedArray((nmr_distributions, nmr_samples), ctype),
                            '_rng_state': KernelArray(rng_state, 'uint')})

        runner = RunProcedure(self._cl_runtime_info)
        runner.run_procedure(cl_function, kernel_data, nmr_distributions)

        return kernel_data['samples'].get_data()

    def _get_uniform_kernel(self, nmr_samples, ctype):
        random_library = Rand123()
        src = random_library.get_cl_code()
        src += '''
            void compute(mot_data_struct* data){
                rand123_data rand123_rng_data = rand123_initialize_data((uint[]){
                    data->_rng_state[0], data->_rng_state[1], data->_rng_state[2], data->_rng_state[3], 
                    data->_rng_state[4], data->_rng_state[5], 0});
                void* rng_data = (void*)&rand123_rng_data;

                for(uint i = 0; i < ''' + str(nmr_samples) + '''; i++){
                    double4 randomnr = rand4(rng_data);
                    data->samples[i] = (''' + ctype + ''')(data->min_val[0] + 
                                                           randomnr.x * (data->max_val[0] - data->min_val[0]));
                }
            }
        '''
        return NameFunctionTuple('compute', src)

    def _get_gaussian_kernel(self, nmr_samples, ctype):
        random_library = Rand123()
        src = random_library.get_cl_code()
        src += '''
            void compute(mot_data_struct* data){
                rand123_data rand123_rng_data = rand123_initialize_data((uint[]){
                    data->_rng_state[0], data->_rng_state[1], data->_rng_state[2], data->_rng_state[3], 
                    data->_rng_state[4], data->_rng_state[5], 0});
                void* rng_data = (void*)&rand123_rng_data;
                
                for(uint i = 0; i < ''' + str(nmr_samples) + '''; i++){
                    double4 randomnr = randn4(rng_data);
                    data->samples[i] = (''' + ctype + ''')(data->mean[0] + randomnr.x * data->std[0]);
                }
            }
        '''
        return NameFunctionTuple('compute', src)



from pkg_resources import resource_filename

from mot.library_functions import SimpleCLLibraryFromFile, SimpleCLLibrary

__author__ = 'Robbert Harms'
__date__ = '2018-05-12'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class CerfImWOfX(SimpleCLLibraryFromFile):
    def __init__(self):
        """Calculate the cerf."""
        super(CerfImWOfX, self).__init__(
            'void', 'cerf', [],
            resource_filename('mot', 'data/opencl/cerf/im_w_of_x.cl'))


class dawson(SimpleCLLibrary):
    def __init__(self):
        """Calculate the Dawson's integral for a real argument."""
        super(dawson, self).__init__(
            'double', 'dawson', [('double', 'x')],
            '''return (sqrt(M_PI)/2.0) * im_w_of_x(x);''',
            dependencies=(CerfImWOfX(),))


class erfi(SimpleCLLibrary):
    def __init__(self):
        """Calculate the imaginary error function for a real argument (special case).

        Compute erfi(x) = -i erf(ix), the imaginary error function.
        """
        super(erfi, self).__init__('double', 'erfi', [('double', 'x')], '''
            return x*x > 720 ? (x > 0 ? INFINITY : -INFINITY) : exp(x*x) * im_w_of_x(x);
        ''', dependencies=(CerfImWOfX(),))




import os
from pkg_resources import resource_filename
from .base import AbstractParallelOptimizer

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Powell(AbstractParallelOptimizer):

    default_patience = 2

    def __init__(self, reset_method=None, patience=None, optimizer_settings=None, patience_line_search=None, **kwargs):
        """Use the Powell method to calculate the optimum.

        Args:
            patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
            reset_method (str): one of 'EXTRAPOLATED_POINT' or 'RESET_TO_IDENTITY' lower case or upper case.
            patience_line_search (int): the patience of the searching algorithm. Defaults to the
                same patience as for the Powell algorithm itself.
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}

        keyword_values = {}
        keyword_values['reset_method'] = reset_method
        keyword_values['patience_line_search'] = patience_line_search

        option_defaults = {'reset_method': 'EXTRAPOLATED_POINT', 'patience_line_search': patience}

        def get_value(option_name):
            value = keyword_values.get(option_name)
            if value is None:
                value = optimizer_settings.get(option_name)
            if value is None:
                value = option_defaults[option_name]
            return value

        for option in option_defaults:
            optimizer_settings.update({option: get_value(option)})

        super(Powell, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def _get_optimization_function(self, model):
        params = {'NMR_PARAMS': model.get_nmr_parameters(), 'PATIENCE': self.patience}

        for option, value in self._optimizer_settings.items():
            params.update({option.upper(): value})
        params['RESET_METHOD'] = 'POWELL_RESET_METHOD_' + params['RESET_METHOD'].upper()

        powell_code = open(os.path.abspath(resource_filename('mot', 'data/opencl/powell.cl')), 'r').read()
        if params:
            powell_code = powell_code % params
        return powell_code

    def _get_optimizer_call_name(self):
        return 'powell'



from mot.cl_runtime_info import CLRuntimeInfo

__author__ = 'Robbert Harms'
__date__ = "2014-04-26"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CLRoutine(object):

    def __init__(self, cl_runtime_info=None):
        """Base class for CL routines.

        Args:
            cl_runtime_info (mot.cl_runtime_info.CLRuntimeInfo): the runtime information
        """
        self._cl_runtime_info = cl_runtime_info or CLRuntimeInfo()

    def set_cl_runtime_info(self, cl_runtime_info):
        """Update the CL runtime information.

        Args:
            cl_runtime_info (mot.cl_runtime_info.CLRuntimeInfo): the new runtime information
        """
        self._cl_runtime_info = cl_runtime_info

    # todo remove in future versions
    def _create_workers(self, worker_generating_cb):
        """Create workers for all the CL environments in current use.

        Args:
            worker_generating_cb (python function): the callback function that we use to generate the
                worker for a specific CL environment. This should accept as single argument a CL environment and
                should return a Worker instance for use in CL computations.
        """
        cl_environments = self._cl_runtime_info.get_cl_environments()
        return [worker_generating_cb(env) for env in cl_environments]
